@@ -4,7 +4,10 @@ use std::{
 	path::{Path, PathBuf},
 };
 
-use super::fs::{FsBackend, FsError};
+use super::{
+	fs::{FsBackend, FsError},
+	FsErrorType,
+};
 use crate::Entry;
 
 /// A JSON based backend.
@@ -18,12 +21,12 @@ impl JsonBackend {
 	///
 	/// # Errors
 	///
-	/// Returns a [`FsError::PathNotDirectory`] if the given path is not a directory.
+	/// Returns an [`FsError`] if the given path is not a directory.
 	pub fn new<P: AsRef<Path>>(path: P) -> Result<Self, FsError> {
 		let path = path.as_ref().to_path_buf();
 
 		if path.is_file() {
-			Err(FsError::PathNotDirectory(path))
+			Err(FsError::path_not_directory(path))
 		} else {
 			Ok(Self(path))
 		}
@@ -39,14 +42,20 @@ impl FsBackend for JsonBackend {
 		R: io::Read,
 		T: Entry,
 	{
-		serde_json::from_reader(rdr).map_err(|_| FsError::Serde)
+		serde_json::from_reader(rdr).map_err(|e| FsError {
+			source: Some(Box::new(e)),
+			kind: FsErrorType::Deserialization,
+		})
 	}
 
 	fn to_bytes<T>(value: &T) -> Result<Vec<u8>, FsError>
 	where
 		T: Entry,
 	{
-		serde_json::to_vec(value).map_err(|_| FsError::Serde)
+		serde_json::to_vec(value).map_err(|e| FsError {
+			source: Some(Box::new(e)),
+			kind: FsErrorType::Serialization,
+		})
 	}
 
 	fn base_directory(&self) -> PathBuf {
@@ -65,12 +74,12 @@ impl JsonPrettyBackend {
 	///
 	/// # Errors
 	///
-	/// Returns a [`FsError::PathNotDirectory`] if the given path is not a directory.
+	/// Returns an [`FsError`] if the given path is not a directory.
 	pub fn new<P: AsRef<Path>>(path: P) -> Result<Self, FsError> {
 		let path = path.as_ref().to_path_buf();
 
 		if path.is_file() {
-			Err(FsError::PathNotDirectory(path))
+			Err(FsError::path_not_directory(path))
 		} else {
 			Ok(Self(path))
 		}
@@ -86,14 +95,20 @@ impl FsBackend for JsonPrettyBackend {
 		R: io::Read,
 		T: Entry,
 	{
-		serde_json::from_reader(rdr).map_err(|_| FsError::Serde)
+		serde_json::from_reader(rdr).map_err(|e| FsError {
+			source: Some(Box::new(e)),
+			kind: FsErrorType::Deserialization,
+		})
 	}
 
 	fn to_bytes<T>(value: &T) -> Result<Vec<u8>, FsError>
 	where
 		T: Entry,
 	{
-		serde_json::to_vec_pretty(value).map_err(|_| FsError::Serde)
+		serde_json::to_vec_pretty(value).map_err(|e| FsError {
+			source: Some(Box::new(e)),
+			kind: FsErrorType::Serialization,
+		})
 	}
 
 	fn base_directory(&self) -> PathBuf {
